"""
Mesh field and statistics (:mod:`~triumvirate._field`)
======================================================

Parse mesh fields and their statistics.

"""
# STYLE: Un-Pythonic import order to prioritise Cython imports.
from cython.operator cimport dereference as deref

import numpy as np
cimport numpy as np

from ._field cimport CppFieldStats
from .dataobjs cimport Binning
from .parameters cimport ParameterSet

np.import_array()


def _record_binned_vectors(Binning binning not None,
                           ParameterSet paramset not None):
    """Record binned vectors.

    Parameters
    ----------
    binning : :class:`~triumvirate.dataobjs.Binning`
        Binning for the measurements.
    paramset : :class:`~triumvirate.parameters.ParameterSet`
        Parameter set for the sampling mesh grid.

    Returns
    -------
    binned_vectors : :class:`numpy.ndarray`
        Binned vectors as a structured array with the following fields---

        - ``'index'``: bin index;
        - ``'lower_edge'``: lower edge of the bin;
        - ``'upper_edge'``: upper edge of the bin;
        - ``'vecx'``: x-component of the vector;
        - ``'vecy'``: y-component of the vector;
        - ``'vecz'``: z-component of the vector,

    """
    fieldstats_ptr = new CppFieldStats(deref(paramset.thisptr), False)

    binned_vectors_struct = fieldstats_ptr.record_binned_vectors(
        deref(binning.thisptr), ''.encode('utf-8')
    )

    binned_vectors = np.empty(
        binned_vectors_struct.count,
        dtype=[
            ('index', 'i4'),
            ('lower_edge', 'f8'),
            ('upper_edge', 'f8'),
            ('vecx', 'f8'),
            ('vecy', 'f8'),
            ('vecz', 'f8'),
        ]
    )

    binned_vectors['index'] = np.asarray(binned_vectors_struct.indices)
    binned_vectors['lower_edge'] = np.asarray(
        binned_vectors_struct.lower_edges
    )
    binned_vectors['upper_edge'] = np.asarray(
        binned_vectors_struct.upper_edges
    )
    binned_vectors['vecx'] = np.asarray(binned_vectors_struct.vecx)
    binned_vectors['vecy'] = np.asarray(binned_vectors_struct.vecy)
    binned_vectors['vecz'] = np.asarray(binned_vectors_struct.vecz)

    return binned_vectors



"""
Two-Point Correlator Algorithms (:mod:`~triumvirate._twopt`)
============================================================

Declare and interface with two-point correlator algorithms.

"""
# STYLE: Un-Pythonic import order to prioritise Cython imports.
from cython.operator cimport dereference as deref
from libc.stdlib cimport free, malloc
from libcpp.string cimport string

import numpy as np
cimport numpy as np

from ._particles cimport CppParticleCatalogue, _ParticleCatalogue
from .dataobjs cimport (
    Binning, CppBinning,
    LineOfSight,
    PowspecMeasurements, TwoPCFMeasurements, TwoPCFWindowMeasurements
)
from .parameters cimport CppParameterSet, ParameterSet

np.import_array()


cdef extern from "include/twopt.hpp":
    # --------------------------------------------------------------------
    # Normalisation
    # --------------------------------------------------------------------

    double calc_powspec_normalisation_from_particles_cpp \
        "trv::calc_powspec_normalisation_from_particles" (
            CppParticleCatalogue& particles,
            double alpha
        ) except +

    double calc_powspec_normalisation_from_mesh_cpp \
        "trv::calc_powspec_normalisation_from_mesh" (
            CppParticleCatalogue& particles,
            CppParameterSet& params,
            double alpha
        )

    double calc_powspec_normalisation_from_meshes_cpp \
        "trv::calc_powspec_normalisation_from_meshes" (
            CppParticleCatalogue& particles_data,
            CppParticleCatalogue& particles_rand,
            CppParameterSet& params,
            double alpha
        )

    double calc_powspec_normalisation_from_meshes_cpp \
        "trv::calc_powspec_normalisation_from_meshes" (
            CppParticleCatalogue& particles_data,
            CppParticleCatalogue& particles_rand,
            CppParameterSet& params,
            double alpha,
            double padding,
            double cellsize,
            string assignment
        )


    # --------------------------------------------------------------------
    # Full correlator
    # --------------------------------------------------------------------

    PowspecMeasurements compute_powspec_cpp "trv::compute_powspec" (
        CppParticleCatalogue& catalogue_data,
        CppParticleCatalogue& catalogue_rand,
        LineOfSight* los_data,
        LineOfSight* los_rand,
        CppParameterSet& params,
        CppBinning& kbinning,
        double norm_factor
    )

    TwoPCFMeasurements compute_corrfunc_cpp "trv::compute_corrfunc" (
        CppParticleCatalogue& catalogue_data,
        CppParticleCatalogue& catalogue_rand,
        LineOfSight* los_data,
        LineOfSight* los_rand,
        CppParameterSet& params,
        CppBinning& rbinning,
        double norm_factor
    )

    PowspecMeasurements compute_powspec_in_gpp_box_cpp \
        "trv::compute_powspec_in_gpp_box" (
            CppParticleCatalogue& catalogue_data,
            CppParameterSet& params,
            CppBinning& kbinning,
            double norm_factor
        )

    TwoPCFMeasurements compute_corrfunc_in_gpp_box_cpp \
        "trv::compute_corrfunc_in_gpp_box" (
            CppParticleCatalogue& catalogue_data,
            CppParameterSet& params,
            CppBinning& rbinning,
            double norm_factor
        )

    TwoPCFWindowMeasurements compute_corrfunc_window_cpp \
        "trv::compute_corrfunc_window" (
            CppParticleCatalogue& catalogue_rand,
            LineOfSight* los_rand,
            CppParameterSet& params,
            CppBinning& rbinning,
            double alpha,
            double norm_factor
        )


def _calc_powspec_normalisation_from_particles(
        _ParticleCatalogue particles not None, double alpha
    ):
    return calc_powspec_normalisation_from_particles_cpp(
        deref(particles.thisptr), alpha
    )


def _calc_powspec_normalisation_from_mesh(
        _ParticleCatalogue particles not None,
        ParameterSet params not None,
        double alpha
    ):
    return calc_powspec_normalisation_from_mesh_cpp(
        deref(particles.thisptr), deref(params.thisptr), alpha
    )


def _calc_powspec_normalisation_from_meshes(
        _ParticleCatalogue particles_data not None,
        _ParticleCatalogue particles_rand not None,
        ParameterSet params not None,
        double alpha,
        padding=None, cellsize=None, assignment=None
    ):
    if None in [padding, cellsize, assignment]:
        return calc_powspec_normalisation_from_meshes_cpp(
            deref(particles_data.thisptr), deref(particles_rand.thisptr),
            deref(params.thisptr), alpha
        )
    else:  # STYLE: non-Pythonic use of `else`
        if not (
            isinstance(padding, float)
            and isinstance(cellsize, float)
            and isinstance(assignment, str)
        ):
            raise TypeError(
                "`padding`, `cellsize` and `assignment` must be "
                "of type float, float and str: received {}, {} and {}."
                .format(type(padding), type(cellsize), type(assignment))
            )
        return calc_powspec_normalisation_from_meshes_cpp(
            deref(particles_data.thisptr), deref(particles_rand.thisptr),
            deref(params.thisptr), alpha,
            padding, cellsize, assignment.encode('utf-8')
        )


def _compute_powspec(
        _ParticleCatalogue catalogue_data not None,
        _ParticleCatalogue catalogue_rand not None,
        np.ndarray[double, ndim=2, mode='c'] los_data not None,
        np.ndarray[double, ndim=2, mode='c'] los_rand not None,
        ParameterSet params not None,
        Binning kbinning not None,
        double norm_factor
    ):
    # Parse lines of sight per particle.
    cdef LineOfSight* los_data_cpp = <LineOfSight*>malloc(
        len(los_data) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_data):
        los_data_cpp[pid].pos[0] = los_x
        los_data_cpp[pid].pos[1] = los_y
        los_data_cpp[pid].pos[2] = los_z

    cdef LineOfSight* los_rand_cpp = <LineOfSight*>malloc(
        len(los_rand) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_rand):
        los_rand_cpp[pid].pos[0] = los_x
        los_rand_cpp[pid].pos[1] = los_y
        los_rand_cpp[pid].pos[2] = los_z

    # Run algorithm.
    cdef PowspecMeasurements results
    results = compute_powspec_cpp(
        deref(catalogue_data.thisptr), deref(catalogue_rand.thisptr),
        los_data_cpp, los_rand_cpp,
        deref(params.thisptr), deref(kbinning.thisptr),
        norm_factor
    )

    free(los_data_cpp); free(los_rand_cpp)

    return {
        'kbin': np.asarray(results.kbin),
        'keff': np.asarray(results.keff),
        'nmodes': np.asarray(results.nmodes),
        'pk_raw': np.asarray(results.pk_raw),
        'pk_shot': np.asarray(results.pk_shot),
    }


def _compute_corrfunc(
        _ParticleCatalogue catalogue_data not None,
        _ParticleCatalogue catalogue_rand not None,
        np.ndarray[double, ndim=2, mode='c'] los_data not None,
        np.ndarray[double, ndim=2, mode='c'] los_rand not None,
        ParameterSet params not None,
        Binning rbinning not None,
        double norm_factor
    ):
    # Parse lines of sight per particle.
    cdef LineOfSight* los_data_cpp = <LineOfSight*>malloc(
        len(los_data) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_data):
        los_data_cpp[pid].pos[0] = los_x
        los_data_cpp[pid].pos[1] = los_y
        los_data_cpp[pid].pos[2] = los_z

    cdef LineOfSight* los_rand_cpp = <LineOfSight*>malloc(
        len(los_rand) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_rand):
        los_rand_cpp[pid].pos[0] = los_x
        los_rand_cpp[pid].pos[1] = los_y
        los_rand_cpp[pid].pos[2] = los_z

    # Run algorithm.
    cdef TwoPCFMeasurements results
    results = compute_corrfunc_cpp(
        deref(catalogue_data.thisptr), deref(catalogue_rand.thisptr),
        los_data_cpp, los_rand_cpp,
        deref(params.thisptr), deref(rbinning.thisptr),
        norm_factor
    )

    free(los_data_cpp); free(los_rand_cpp)

    return {
        'rbin': np.asarray(results.rbin),
        'reff': np.asarray(results.reff),
        'npairs': np.asarray(results.npairs),
        'xi': np.asarray(results.xi),
    }


def _compute_powspec_in_gpp_box(
        _ParticleCatalogue catalogue_data not None,
        ParameterSet params not None,
        Binning kbinning not None,
        double norm_factor
    ):
    cdef PowspecMeasurements results
    results = compute_powspec_in_gpp_box_cpp(
        deref(catalogue_data.thisptr),
        deref(params.thisptr), deref(kbinning.thisptr),
        norm_factor
    )

    return {
        'kbin': np.asarray(results.kbin),
        'keff': np.asarray(results.keff),
        'nmodes': np.asarray(results.nmodes),
        'pk_raw': np.asarray(results.pk_raw),
        'pk_shot': np.asarray(results.pk_shot),
    }


def _compute_corrfunc_in_gpp_box(
        _ParticleCatalogue catalogue_data not None,
        ParameterSet params not None,
        Binning rbinning not None,
        double norm_factor
    ):
    cdef TwoPCFMeasurements results
    results = compute_corrfunc_in_gpp_box_cpp(
        deref(catalogue_data.thisptr),
        deref(params.thisptr), deref(rbinning.thisptr),
        norm_factor
    )

    return {
        'rbin': np.asarray(results.rbin),
        'reff': np.asarray(results.reff),
        'npairs': np.asarray(results.npairs),
        'xi': np.asarray(results.xi),
    }


def _compute_corrfunc_window(
        _ParticleCatalogue catalogue_rand not None,
        np.ndarray[double, ndim=2, mode='c'] los_rand not None,
        ParameterSet params not None,
        Binning rbinning not None,
        double alpha,
        double norm_factor
    ):
    # Parse lines of sight per particle.
    cdef LineOfSight* los_rand_cpp = <LineOfSight*>malloc(
        len(los_rand) * sizeof(LineOfSight)
    )
    for pid, (los_x, los_y, los_z) in enumerate(los_rand):
        los_rand_cpp[pid].pos[0] = los_x
        los_rand_cpp[pid].pos[1] = los_y
        los_rand_cpp[pid].pos[2] = los_z

    # Run algorithm.
    cdef TwoPCFWindowMeasurements results
    results = compute_corrfunc_window_cpp(
        deref(catalogue_rand.thisptr), los_rand_cpp,
        deref(params.thisptr), deref(rbinning.thisptr),
        alpha, norm_factor
    )

    free(los_rand_cpp)

    return {
        'rbin': np.asarray(results.rbin),
        'reff': np.asarray(results.reff),
        'npairs': np.asarray(results.npairs),
        'xi': np.asarray(results.xi),
    }



"""Test :mod:`~triumvirate._arrayops`.

"""
import warnings

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal, assert_raises

from triumvirate._arrayops import (
    _check_1d_array,
    _check_2d_array,
    ShapeError,
    OrderError,
    MixedSignError,
    SpacingError,
    extrap_pad,
    extrap_lin,
    extrap_loglin,
    extrap_loglin_oscil,
    extrap2d_lin,
    extrap2d_loglin,
    extrap2d_pad,
    extrap2d_loglin_oscil,
    reshape_threept_datatab,
)


@pytest.mark.parametrize(
    "a, check_sorted, check_epsign, check_lin, check_loglin, exception",
    [
        # Test case 1: check sorted array
        (
            np.array([1., 3., 2.,]), True, False, False, False,
            OrderError,
        ),
        # Test case 2: check mixed sign entries at endpoints
        (
            np.array([1., -2., 3.,]), False, True, False, False,
            MixedSignError,
        ),
        # Test case 3: check linearly spaced array
        (
            np.array([1., 2., 4.,]), False, False, True, False,
            SpacingError,
        ),
        # Test case 4: check log-linearly spaced array
        (
            np.array([1., 2., 3.,]), False, False, False, True,
            SpacingError,
        ),
        # Test case 5: check array dimension
        (
            np.array([[1., 2., 3.,]]*2), False, False, False, False,
            ShapeError,
        ),
    ],
)
def test__check_1d_array(a,
                         check_sorted, check_epsign, check_lin, check_loglin,
                         exception):
    if exception is not None:
        with pytest.raises(exception):
            a_out = _check_1d_array(
                a, check_sorted, check_epsign, check_lin, check_loglin
            )
        return

    a_out = _check_1d_array(
        a, check_sorted, check_epsign, check_lin, check_loglin
    )
    assert_allclose(a_out, a)


@pytest.mark.parametrize(
    "a, check_sorted, check_epsign, check_lin, check_loglin, exception",
    [
        # Test case 1: check sorted array
        (
            np.array([[1., 2., 3.,], [2., 3., 2.,]]),
            True, False, False, False, OrderError,
        ),
        # Test case 2: check mixed sign entries at endpoints
        (
            np.array([[1., 2., 3.,], [2., -3., 2.,]]),
            False, True, False, False, MixedSignError,
        ),
        # Test case 3: check linearly spaced array
        (
            np.array([[1., 2., 3.,], [1., 2., 4.,]]),
            False, False, True, False, SpacingError,
        ),
        # Test case 4: check log-linearly spaced array
        (
            np.array([[1., 2., 3.,], [1., 2., 4.,]]),
            False, False, False, True, SpacingError,
        ),
        # Test case 5: check array dimension
        (
            np.array([1, 2, 3,]), False, False, False, False, ShapeError,
        ),
    ],
)
def test__check_2d_array(a,
                         check_sorted, check_epsign, check_lin, check_loglin,
                         exception):
    if exception is not None:
        with pytest.raises(exception):
            a_out = _check_2d_array(
                a, check_sorted, check_epsign, check_lin, check_loglin
            )
        return

    a_out = _check_2d_array(
        a, check_sorted, check_epsign, check_lin, check_loglin
    )
    assert_allclose(a_out, a)


@pytest.mark.parametrize(
    "a, n_ext, c_lower, c_upper, expected_output",
    [
        # Test case 1: extrapolate with positive padding
        (
            np.array([1, 2, 3,]), 2, -1, 1,
            np.array([-1, -1, 1, 2, 3, 1, 1,]),
        ),
        # Test case 2: extrapolate with no padding
        (
            np.array([1, 2, 3,]), 0, 0, 0,
            np.array([1, 2, 3,]),
        ),
    ],
)
def test_extrap_pad(a, n_ext, c_lower, c_upper, expected_output):
    a_out = extrap_pad(a, n_ext, c_lower, c_upper)
    assert_allclose(a_out, expected_output)


@pytest.mark.parametrize(
    "a, n_ext, expected_output",
    [
        # Test case 1: extrapolate
        (
            np.array([1, 2, 2,]), 2,
            np.array([-1, 0, 1, 2, 2, 2, 2,]),
        ),
        # Test case 2: extrapolate not
        (
            np.array([1, 2, 3,]), 0,
            np.array([1, 2, 3,]),
        ),
    ],
)
def test_extrap_lin(a, n_ext, expected_output):
    a_out = extrap_lin(a, n_ext)
    assert_allclose(a_out, expected_output)


@pytest.mark.parametrize(
    "a, n_ext, expected_output",
    [
        # Test case 1: extrapolate log-linearly
        (
            np.array([1., 2., 3.,]), 1,
            np.array([.5, 1., 2., 3., 4.5,]),
        ),
        # Test case 2: extrapolate log-linearly with negative values
        (
            np.array([-1., -2., -3.,]), 1,
            np.array([-.5, -1., -2., -3., -4.5,]),
        ),
        # Test case 3: extrapolate log-linearly with zero values
        (
            np.array([0, 1, 2,]), 2,
            np.array([0, 0, 0, 1, 2, 4, 8,]),
        ),
    ],
)
def test_extrap_loglin(a, n_ext, expected_output):
    a_out = extrap_loglin(a, n_ext)
    assert_allclose(a_out, expected_output)


@pytest.mark.parametrize(
    "a, n_ext, expected_output",
    [
        # Test case 1: extrapolate log-linearly
        (
            np.array([1., 2., 4., 8., 16., 8., 4., 2., 1.,]), 1,
            np.array([.5, 1., 2., 4., 8., 16., 8., 4., 2., 1., .5,]),
        ),
        # Test case 2: extrapolate log-linearly with negative values
        (
            - np.array([1., 2., 4., 8., 16., 8., 4., 2., 1.,]), 1,
            - np.array([.5, 1., 2., 4., 8., 16., 8., 4., 2., 1., .5,]),
        ),
    ],
)
def test_extrap_loglin_oscil(a, n_ext, expected_output):
    a_out = extrap_loglin_oscil(a, n_ext)
    assert_allclose(a_out, expected_output)


@pytest.mark.parametrize(
    "a, n_ext,"
    "crow_lower, crow_upper, ccol_lower, ccol_upper, expected_output",
    [
        # Test case 1: extrapolate with positive padding
        (
            np.array([[1, 2, 3,], [4, 5, 6,]]), 1,
            [-1, 1,], [-1, -1,], [-1, -2, -1,], [1, 2, -1,],
            np.array([
                # autopep8: off
                [-1, -1, -2, -1, -1,],
                [-1,  1,  2,  3, -1,],
                [ 1,  4,  5,  6, -1,],  # noqa: E201
                [ 1,  1,  2, -1, -1,],  # noqa: E201
                # autopep8: on
            ]),
        ),
    ],
)
def test_extrap2d_pad(a, n_ext, crow_lower, crow_upper, ccol_lower, ccol_upper,
                      expected_output):
    a_out = extrap2d_pad(
        a, n_ext, crow_lower, crow_upper, ccol_lower, ccol_upper
    )
    assert_allclose(a_out, expected_output)


@pytest.mark.parametrize(
    "a, n_ext, expected_output",
    [
        # Test case 1: extrapolate bilinearly
        (
            np.array([[1, 2, 3,], [4, 5, 6,]]), 1,
            np.array([
                # autopep8: off
                [-3, -2, -1,  0,  1,],
                [ 0,  1,  2,  3,  4,],  # noqa: E201
                [ 3,  4,  5,  6,  7,],  # noqa: E201
                [ 6,  7,  8,  9, 10,],  # noqa: E201
                # autopep8: on
            ]),
        ),
    ],
)
def test_extrap2d_lin(a, n_ext, expected_output):
    a_out = extrap2d_lin(a, n_ext)
    assert_allclose(a_out, expected_output)


@pytest.mark.parametrize(
    "a, n_ext, expected_output",
    [
        # Test case 1: Extrapolate log-linearly
        (
            np.array([[1., 2., 4.,], [2., 4., 8.,]]), 1,
            np.array([
                # autopep8: off
                [ .25,  .5, 1.,  2.,  4.,],  # noqa: E201
                [ .5 , 1. , 2.,  4.,  8.,],  # noqa: E201, E203
                [1.  , 2. , 4.,  8., 16.,],  # noqa: E201, E203
                [2.  , 4. , 8., 16., 32.,],  # noqa: E201, E203
                # autopep8: on
            ]),
        ),
    ],
)
def test_extrap2d_loglin(a, n_ext, expected_output):
    a_out = extrap2d_loglin(a, n_ext)
    assert_allclose(a_out, expected_output)


@pytest.mark.parametrize(
    "a, n_ext, expected_output",
    [
        # Test case 1: Extrapolate log-linearly with oscillatory behavior
        (
            np.array([
                # autopep8: off
                [1. ,  2.,  4.,  8.,  4.,  2., 1.,],  # noqa: E203
                [2. ,  4.,  8., 16.,  8.,  4., 2.,],  # noqa: E203
                [4. ,  8., 16., 32., 16.,  8., 4.,],  # noqa: E203
                [8. , 16., 32., 64., 32., 16., 8.,],  # noqa: E203
                [4. ,  8., 16., 32., 16.,  8., 4.,],  # noqa: E203
                [2. ,  4.,  8., 16.,  8.,  4., 2.,],  # noqa: E203
                [1. ,  2.,  4.,  8.,  4.,  2., 1.,],  # noqa: E203
                # autopep8: on
            ]),
            1,
            np.array([
                # autopep8: off
                [ .25,  .5,  1.,  2.,  4.,  2.,  1.,  .5,  .25,],  # noqa: E201
                [ .5 , 1. ,  2.,  4.,  8.,  4.,  2., 1. ,  .5 ,],  # noqa: E201, E203, E501
                [1.  , 2. ,  4.,  8., 16.,  8.,  4., 2. , 1.  ,],  # noqa: E201, E203, E501
                [2.  , 4. ,  8., 16., 32., 16.,  8., 4. , 2.  ,],  # noqa: E201, E203, E501
                [4.  , 8. , 16., 32., 64., 32., 16., 8. , 4.  ,],  # noqa: E201, E203, E501
                [2.  , 4. ,  8., 16., 32., 16.,  8., 4. , 2.  ,],  # noqa: E201, E203, E501
                [1.  , 2. ,  4.,  8., 16.,  8.,  4., 2. , 1.  ,],  # noqa: E201, E203, E501
                [ .5 , 1. ,  2.,  4.,  8.,  4.,  2., 1. ,  .5 ,],  # noqa: E201, E203, E501
                [ .25,  .5,  1.,  2.,  4.,  2.,  1.,  .5,  .25,],  # noqa: E201
                # autopep8: on
            ])
        )
    ]
)
def test_extrap2d_loglin_oscil(a, n_ext, expected_output):
    warnings.filterwarnings(
        "ignore", category=UserWarning,
        message="Array length is less than 8. "
                "Extrapolation may not be accurate."
    )
    a_out = extrap2d_loglin_oscil(a, n_ext)
    assert_allclose(a_out, expected_output)


@pytest.mark.parametrize(
    "paired_coords, flat_multipole, shape,"
    "expected_common_coords, expected_multipole",
    [
        # Test case 1: Valid input
        (
            np.array([[0, 0], [0, 1], [1, 0]]),
            np.array([1, 2, 3]),
            'triu',
            np.array([0, 1,]),
            np.array([[1, 2,], [2, 3],]),
        ),
        # Test case 2: Invalid input - paired_coords is not a 2-d array
        (
            np.array([1, 2, 3,]),
            np.array([1, 2, 3,]),
            None,
            None,
            None,
        ),
        # Test case 3: Invalid input - unmatching dimensions
        (
            np.array([[1, 2], [2, 3], [3, 4]]),
            np.array([1, 2, 3, 4,]),
            None,
            None,
            None,
        ),
        # Test case 4: Invalid input - not an upper triangular matrix
        (
            np.array([[1, 2], [2, 3], [3, 4]]),
            np.array([1, 2, 3, 4, 5, 6,]),
            None,
            None,
            None,
        ),
        # Test case 5: Valid input
        (
            np.array([[0, 0], [0, 1], [1, 0], [1, 1]]),
            np.array([1, 2, 3, 4,]),
            'full',
            np.array([0, 1]),
            np.array([[1, 2,], [3, 4,],])
        ),
    ]
)
def test_reshape_threept_datatab(paired_coords, flat_multipole, shape,
                                 expected_common_coords, expected_multipole):
    if expected_common_coords is not None and expected_multipole is not None:
        common_coords, multipole = reshape_threept_datatab(
            paired_coords, flat_multipole, shape=shape
        )
        assert_array_equal(common_coords, expected_common_coords)
        assert_array_equal(multipole, expected_multipole)
    else:
        assert_raises(
            ValueError, reshape_threept_datatab, paired_coords, flat_multipole
        )



"""
##########################################################################
``Triumvirate`` | Three-Point Clustering Measurements in LSS
##########################################################################

``Triumvirate`` is a Python/C++ package for measuring three-point (and
two-point) clustering statistics in large-scale structure (LSS)
cosmological analyses.

.. topic:: GNU General Public License, version 3

    Copyright (C) 2023, Mike S Wang & Naonori S Sugiyama

    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation, either version 3 of the License, or (at your
    option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program.  If not, see `<https://www.gnu.org/licenses/>`_.

"""  # numpydoc ignore=SS03
from importlib.metadata import PackageNotFoundError, version

# Installation validation
from ._valid_install import validate_installation  # noqa: F401


__copyright__ = 'Copyright 2023, Mike S Wang & Naonori S Sugiyama'
__date__ = '2023-03-30'
__license__ = 'GPL-3.0-or-later'

try:
    __version__ = version('triumvirate')
except PackageNotFoundError:
    __version__ = '0.5.0'  # fallback version number



"""
Array Operations (:mod:`~triumvirate._arrayops`)
================================================

Perform array operations.

.. autosummary::
    ShapeError
    OrderError
    MixedSignError
    SpacingError
    DivergenceWarning

1-d arrays:

.. autosummary::
    extrap_pad
    extrap_lin
    extrap_loglin
    extrap_loglin_oscil

2-d arrays:

.. autosummary:
    extrap2d_lin
    extrap2d_loglin
    extrap2d_pad
    extrap2d_loglin_oscil
    reshape_threept_datatab

"""  # numpydoc ignore=SS01
import warnings
from collections.abc import Sequence

import numpy as np


class ShapeError(ValueError):
    """Exception raised when the array shape is incorrect.

    """


class OrderError(ValueError):
    """Exception raised when the array order is incorrect.

    """


class MixedSignError(ValueError):
    """Exception raised when the array signs are mixed.

    """


class SpacingError(ValueError):
    """Exception raised when the array spacing is incorrect.

    """


class DivergenceWarning(RuntimeWarning):
    """Warning issued when divergent behaviour is detected in an array.

    """


# ========================================================================
# Extrapolation
# ========================================================================

def _check_1d_array(a, check_sorted=False, check_epsign=False,
                    check_lin=False, check_loglin=False):
    """Check the input array is a 1-d array, possibly satisfying
    additional conditions.

    Parameters
    ----------
    a : array of float
        Input array.
    check_sorted : bool, optional
        If `True` (default is `False`), check the input array is sorted.
    check_epsign : bool, optional
        If `True` (default is `False`), check for mixed sign entries at
        the endpoints in the input array, and for nonzero endpoints with
        zero next-to-endpoint entries.
    check_lin : bool, optional
        If `True` (default is `False`), check the input array is
        linearly spaced.
    check_loglin : bool, optional
        If `True` (default is `False`), check the input array is
        log-linearly spaced (and all entries have the same sign).

    Returns
    -------
    a : 1-d array of float
        Squeezed 1-d array.

    Raises
    ------
    ShapeError
        When the input array is not equivalently 1-d.
    ValueError
        When the input array fails any of the checks specified.

    """
    a = np.squeeze(a)

    if a.ndim != 1 or a.size < 2:
        raise ShapeError("Input array is not 1-d.")

    def _check_epsign(a, strict=False):  # numpydoc ignore=GL08
        if strict:
            same_sign_l = a[0] * a[1] > 0.
            same_sign_r = a[-1] * a[-2] > 0.
        else:
            same_sign_l = (a[0] * a[1] > 0.) or (a[0] == 0.)
            same_sign_r = (a[-1] * a[-2] > 0.) or (a[-1] == 0.)
        if not (same_sign_l and same_sign_r):
            raise MixedSignError(
                "Input array contains mixed-sign entries at the endpoints."
            )

    if check_sorted:
        if (not np.all(a[:-1] <= a[1:])) or (not np.all(a[:-1] >= a[1:])):
            raise OrderError("Input array is not a sorted sequence.")
    if check_epsign:
        _check_epsign(a)
    if check_lin:
        spacing = np.mean(a[1:] - a[:-1])
        if not np.allclose(a[1:] - a[:-1], spacing):
            raise SpacingError("Input array is not linearly spaced.")
    if check_loglin:
        _check_epsign(a, strict=True)
        log_spacing = np.mean(np.log(a[1:] / a[:-1]))
        if not np.allclose(np.log(a[1:] / a[:-1]), log_spacing):
            raise SpacingError("Input array is not log-linearly spaced.")

    return a


def _check_2d_array(a, check_sorted=False, check_epsign=False,
                    check_lin=False, check_loglin=False):
    """Check the input array is a sorted 2-d array, possibly satisfying
    additional conditions.

    Parameters
    ----------
    a : array of float
        Input array.
    check_sorted : bool, optional
        If `True` (default is `False`), check the input array is sorted.
    check_epsign : bool, optional
        If `True` (default is `False`), check for mixed sign entries at
        the endpoints in the input array, and for nonzero endpoints with
        zero next-to-endpoint entries.
    check_lin : bool, optional
        If `True` (default is `False`), check the input array is
        linearly spaced.
    check_loglin : bool, optional
        If `True` (default is `False`), check the input array is
        log-linearly spaced (and all entries have the same sign).

    Returns
    -------
    a : 2-d array of float
        Squeezed 2-d array.

    Raises
    ------
    ShapeError
        When the input array is not equivalently 2-d.
    ValueError
        When the input array fails any of the checks specified.

    """
    a = np.squeeze(a)

    if a.ndim != 2 or a.size < 4:
        raise ShapeError("Input array is not 2-d.")

    def _check_epsign(a, strict=False):  # numpydoc ignore=GL08
        if strict:
            same_sign_l = np.all(a[:, 0] * a[:, 1] > 0.)
            same_sign_r = np.all(a[:, -1] * a[:, -2] > 0.)
            same_sign_u = np.all(a[0, :] * a[1, :] > 0.)
            same_sign_d = np.all(a[-1, :] * a[-2, :] > 0.)
        else:
            same_sign_l = np.all(np.logical_or(
                a[:, 0] * a[:, 1] > 0., a[:, 0] == 0.
            ))
            same_sign_r = np.all(np.logical_or(
                a[:, -1] * a[:, -2] > 0., a[:, -1] == 0.
            ))
            same_sign_u = np.all(np.logical_or(
                a[0, :] * a[1, :] > 0., a[0, :] == 0.
            ))
            same_sign_d = np.all(np.logical_or(
                a[-1, :] * a[-2, :] > 0., a[-1, :] == 0.
            ))

        if not (same_sign_l and same_sign_r and same_sign_u and same_sign_d):
            raise MixedSignError(
                "Input array contains mixed-sign entries at the endpoints."
            )

    if check_sorted:
        for a_, direction in zip([a, a.transpose()], ['row', 'column']):
            if (
                (
                    not np.all([
                        np.all(a__row[:-1] <= a__row[1:]) for a__row in a_
                    ])
                ) or (
                    not np.all([
                        np.all(a__row[:-1] >= a__row[1:]) for a__row in a_
                    ])
                )
            ):
                raise OrderError(
                    "Input array is not a sorted sequence "
                    f"along the {direction}s."
                )
    if check_epsign:
        _check_epsign(a)
    if check_lin:
        for a_, direction in zip([a, a.transpose()], ['row', 'column']):
            spacing = np.mean(a_[:, 1:] - a_[:, :-1], axis=1)
            if not np.allclose(a_[:, 1:] - a_[:, :-1], spacing[:, None]):
                raise SpacingError(
                    "Input array is not linearly spaced "
                    f"along the {direction}s."
                )
    if check_loglin:
        _check_epsign(a, strict=True)
        for a_, direction in zip([a, a.transpose()], ['row', 'column']):
            log_spacing = np.mean(np.log(a_[:, 1:] / a_[:, :-1]), axis=1)
            if not np.allclose(
                np.log(a_[:, 1:] / a_[:, :-1]), log_spacing[:, None]
            ):
                raise SpacingError(
                    "Input array is not log-linearly spaced "
                    f"along the {direction}s."
                )

    return a


def extrap_pad(a, n_ext, c_lower, c_upper):
    """Extrapolate a 1-d array by constant padding.

    Parameters
    ----------
    a : array of float
        Input 1-d array.
    n_ext : int
        Number of extra elements on either side.
    c_lower, c_upper : float
        Constant value used for padding on either side.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 1-d array.

    """
    a = _check_1d_array(a)

    a_l = np.full(n_ext, c_lower)
    a_r = np.full(n_ext, c_upper)
    a_out = np.concatenate([a_l, a, a_r])

    return a_out


def extrap_lin(a, n_ext):
    """Extrapolate a 1-d array linearly.

    The linear spacing used is the distance between the two points
    at either end separately.

    Parameters
    ----------
    a : array of float
        Input 1-d array.
    n_ext : int
        Number of extra elements on either side.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 1-d array.

    """
    a = _check_1d_array(a)

    a_l = a[0] + (a[1] - a[0]) * np.arange(-n_ext, 0)
    a_r = a[-1] + (a[-1] - a[-2]) * np.arange(1, n_ext + 1)
    a_out = np.concatenate([a_l, a, a_r])

    return a_out


def extrap_loglin(a, n_ext):
    """Extrapolate a 1-d array log-linearly.

    The log-linear spacing used is the log-distance between the two points
    at either end separately.

    Parameters
    ----------
    a : array of float
        Input 1-d array.
    n_ext : int
        Number of extra elements on either side.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 1-d array.

    """
    a = _check_1d_array(a, check_epsign=True)

    a_l = a[0] * (a[0] / a[1]) ** np.arange(n_ext, 0, -1)
    a_r = a[-1] * (a[-1] / a[-2]) ** np.arange(1, n_ext + 1, 1)
    a_out = np.concatenate([a_l, a, a_r])

    return a_out


def extrap_loglin_oscil(a, n_ext):
    """Extrapolate a 1-d array with oscillatory behaviour log-linearly.

    Parameters
    ----------
    a : array of float
        Input 1-d array.
    n_ext : int
        Number of extra elements on either side.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 1-d array.

    """
    a = _check_1d_array(a)

    if len(a) < 8:
        warnings.warn(
            "Array length is less than 8. Extrapolation may not be accurate.",
            UserWarning
        )

    # Check for sign changes and restrict extrapolation to a range
    # with at least 2 points on either 'bank' of the 'zero river'.
    signdiff = np.diff(np.sign(a))
    signchange = np.nonzero(signdiff)[0]
    if len(signchange) > 0:
        idx_lbank = max(signchange[0] + 1, 3)
        idx_rbank = min(signchange[-1] + 1, a.size - 4)
    else:
        idx_lbank = a.size - 1
        idx_rbank = 0

    # Check for oscillations and further restrict extrapolation to a range
    # excluding the innermost peak/through with at least 2 points on either
    # 'bank' of the 'zero river'.
    lngrad_lbank = np.nan_to_num(np.gradient(np.log(np.abs(a[:idx_lbank]))))
    signchange_lngrad_lbank = np.nonzero(np.diff(np.sign(lngrad_lbank)))[0]
    if len(signchange_lngrad_lbank) > 0:
        idx_ldbank = max(signchange_lngrad_lbank[-1] + 1, 3)
    else:
        idx_ldbank = a.size - 1

    lngrad_rbank = np.nan_to_num(np.gradient(np.log(np.abs(a[idx_rbank:]))))
    signchange_lngrad_rbank = np.nonzero(np.diff(np.sign(lngrad_rbank)))[0]
    if len(signchange_lngrad_rbank) > 0:
        idx_rdbank = min(idx_rbank + signchange_lngrad_rbank[0], a.size - 4)
    else:
        idx_rdbank = 0

    # Divide the array into two 'halves' either side of the peak with
    # each 'half' halved into 'quarters'.
    idx_peak = np.argmax(np.abs(a))
    if idx_peak == 0 or idx_peak == a.size - 1:
        idx_peak = a.size // 2
    idx_lquater = max((idx_peak + 1) // 2 - 1, 3)
    idx_rquater = min(idx_peak + (a.size - 1 - idx_peak) // 2, a.size - 4)

    # Define the tails used as the extrapolation range.
    idx_ltail = min(idx_lbank, idx_ldbank, idx_lquater)
    idx_rtail = max(idx_rbank, idx_rdbank, idx_rquater)

    # Extrapolate the left and right tails.
    lngrad_ltail = np.nan_to_num(np.gradient(np.log(np.abs(a[:idx_ltail]))))
    # Oscillatory behaviour
    if np.diff(np.sign(lngrad_ltail)).any():
        # Use mean gradient.
        if (mean_lngrad_ltail := np.mean(lngrad_ltail)) >= 0.:
            powlaw_ltail = - mean_lngrad_ltail
            # print("Left tail extrap: trend")
        # Use constant padding when mean gradient is divergent.
        else:
            powlaw_ltail = 0.
            warnings.warn(
                "Divergent behaviour detected at the left tail. "
                "Constant padding is used instead.",
                DivergenceWarning
            )
            # print("Left tail extrap: trend -> const")
    # No oscillations
    else:
        # Use endpoint gradient.
        if (
            lngrad_lend := np.nan_to_num(
                np.log(np.abs(a[1])) - np.log(np.abs(a[0]))
            )
        ) >= 0.:
            powlaw_ltail = - lngrad_lend
            # print("Left tail extrap: endpoint")
        # Endpoint gradient is divergent possibly with sign changes.
        else:
            mean_lngrad_ltail = np.nan_to_num(np.mean(
                np.gradient(np.log(np.abs(a[:idx_lquater])))
            ))
            # Use mean gradient.
            if mean_lngrad_ltail >= 0.:
                powlaw_ltail = - mean_lngrad_ltail
                # print("Left tail extrap: endpoint -> trend")
            # Use constant padding when mean gradient is divergent.
            else:
                powlaw_ltail = 0.
                warnings.warn(
                    "Divergent behaviour detected at the left endpoints. "
                    "Constant padding is used instead.",
                    DivergenceWarning
                )
                # print("Left tail extrap: endpoint -> const")

    lngrad_rtail = np.nan_to_num(np.gradient(np.log(np.abs(a[idx_rtail:]))))
    if np.diff(np.sign(lngrad_rtail)).any():
        if (mean_lngrad_rtail := np.mean(lngrad_rtail)) <= 0.:
            powlaw_rtail = mean_lngrad_rtail
            # print("Right tail extrap: trend")
        else:
            powlaw_rtail = 0.
            warnings.warn(
                "Divergent behaviour detected at the right tail. "
                "Constant padding is used instead.",
                DivergenceWarning
            )
            # print("Right tail extrap: trend -> const")
    else:
        if (
            lngrad_rend := np.nan_to_num(
                np.log(np.abs(a[-1])) - np.log(np.abs(a[-2]))
            )
        ) <= 0.:
            powlaw_rtail = lngrad_rend
            # print("Right tail extrap: endpoint")
        else:
            mean_lngrad_rtail = np.nan_to_num(np.mean(
                np.gradient(np.log(np.abs(a[idx_rquater:])))
            ))
            if mean_lngrad_rtail <= 0.:
                powlaw_rtail = mean_lngrad_rtail
                # print("Right tail extrap: endpoint -> trend")
            else:
                powlaw_rtail = 0.
                warnings.warn(
                    "Divergent behaviour detected at the right endpoints. "
                    "Constant padding is used instead.",
                    DivergenceWarning
                )
                # print("Right tail extrap: endpoint -> const")

    a_l = a[0] * np.exp(powlaw_ltail * np.arange(n_ext, 0, -1))
    a_r = a[-1] * np.exp(powlaw_rtail * np.arange(1, n_ext + 1, 1))
    a_out = np.concatenate([a_l, a, a_r])

    return a_out


def _extrap2d_row_lin(a, ncol_ext):
    """Extrapolate a 2-d array linearly along each row
    (i.e. horizontally).

    Parameters
    ----------
    a : 2-d :class:`numpy.ndarray`
        Input 2-d array.
    ncol_ext : int
        Number of extra columns on either side.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 2-d array.

    """
    a_l = a[:, [0]] + np.arange(-ncol_ext, 0) * (a[:, [1]] - a[:, [0]])
    a_r = a[:, [-1]] + np.arange(1, ncol_ext + 1) * (a[:, [-1]] - a[:, [-2]])

    a_out = np.c_[a_l, a, a_r]

    return a_out


def _extrap2d_row_loglin(a, ncol_ext):
    """Extrapolate a 2-d array log-linearly along each row
    (i.e. horizontally).

    Parameters
    ----------
    a : 2-d :class:`numpy.ndarray`
        Input 2-d array.
    ncol_ext : int
        Number of extra columns on either side.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 2-d array.

    """
    a_l = a[:, [0]] * (a[:, [0]]/a[:, [1]]) ** np.arange(ncol_ext, 0, -1)
    a_r = a[:, [-1]] * (a[:, [-1]]/a[:, [-2]]) ** np.arange(1, ncol_ext + 1, 1)

    a_out = np.c_[a_l, a, a_r]

    return a_out


def _extrap2d_row_loglin_oscil(a, ncol_ext):
    """Extrapolate a 2-d array with oscillatory behaviour log-linearly
    along each row (i.e. horizontally).

    Parameters
    ----------
    a : 2-d :class:`numpy.ndarray`
        Input 2-d array.
    ncol_ext : int
        Number of extra columns on either side.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 2-d array.

    """
    a_out = []
    for a_row in a:
        a_out.append(extrap_loglin_oscil(a_row, ncol_ext))
    a_out = np.array(a_out)

    return a_out


def extrap2d_pad(a, n_ext, crow_lower, crow_upper, ccol_lower, ccol_upper):
    """Extrapolate a 2-d array by constant padding bi-directionally.

    Parameters
    ----------
    a : array of float
        Input 2-d array.
    n_ext : int or tuple of int
        Number of extra elements on either side of each row and/or column.
    crow_lower, crow_upper, ccol_lower, ccol_upper : (sequence of) float
        Constant values used for padding on either side of
        each row and/or column.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 2-d array.

    Notes
    -----
    The padded array will have the following structure schematically:

    .. code-block:: none

        ... ...             ...            ...  ...             ...             ...
        ... ...             ccol_lower[0]  ...  ccol_lower[-1]  ...             ...
        ... crow_lower[0]   a[0, 0]        ...  a[0, -1]        crow_upper[0]   ...
        ... ...             ...            ...  ...             ...             ...
        ... crow_lower[-1]  a[-1, 0]       ...  a[-1, -1]       crow_upper[-1]  ...
        ... ...             ccol_upper[0]  ...  ccol_upper[-1]  ...             ...
        ... ...             ...            ...  ...             ...             ...

    If any of `crow_lower`, `crow_upper`, `ccol_lower` and `ccol_upper`
    is a scalar, it is repeated to form a 1-d array of the
    appropriate length.

    For consistency, it is required that
    ``crow_lower[0] == ccol_lower[0]``,
    ``crow_upper[0] == ccol_lower[-1]``,
    ``crow_lower[-1] == ccol_upper[0]``, and
    ``crow_upper[-1] == ccol_upper[-1]``.

    """  # noqa: E501
    a = _check_2d_array(a)

    if np.isscalar(crow_lower):
        crow_lower = np.full(a.shape[0], crow_lower)
    if np.isscalar(crow_upper):
        crow_upper = np.full(a.shape[0], crow_upper)
    if np.isscalar(ccol_lower):
        ccol_lower = np.full(a.shape[-1], ccol_lower)
    if np.isscalar(ccol_upper):
        ccol_upper = np.full(a.shape[-1], ccol_upper)

    if crow_lower[0] != ccol_lower[0]:
        raise ValueError(
            "Inconsistent lower padding values for rows and columns."
        )
    if crow_upper[0] != ccol_lower[-1]:
        raise ValueError(
            "Inconsistent lower and upper padding values for rows and columns."
        )
    if crow_lower[-1] != ccol_upper[0]:
        raise ValueError(
            "Inconsistent upper and lower padding values for rows and columns."
        )
    if crow_upper[-1] != ccol_upper[-1]:
        raise ValueError(
            "Inconsistent upper padding values for rows and columns."
        )

    nrow_ext, ncol_ext = \
        n_ext if isinstance(n_ext, Sequence) else (n_ext, n_ext)

    a_l = np.tile(crow_lower, (nrow_ext, 1)).T
    a_r = np.tile(crow_upper, (nrow_ext, 1)).T
    a_ = np.c_[a_l, a, a_r]

    ccol_lower = np.concatenate([a_l[0], ccol_lower, a_r[0]])
    ccol_upper = np.concatenate([a_l[-1], ccol_upper, a_r[-1]])

    a__u = np.tile(ccol_lower, (ncol_ext, 1))
    a__d = np.tile(ccol_upper, (ncol_ext, 1))
    a_out = np.r_[a__u, a_, a__d]

    return a_out


def extrap2d_lin(a, n_ext):
    """Extrapolate a 2-d array bilinearly.

    Parameters
    ----------
    a : array of float
        Input 2-d array.
    n_ext : int or tuple of int
        Number of extra elements on either side of each row and/or column.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 2-d array.

    """
    a = _check_2d_array(a)

    nrow_ext, ncol_ext = \
        n_ext if isinstance(n_ext, Sequence) else (n_ext, n_ext)

    # Extrapolate horizontally then vertically.
    a_ = _extrap2d_row_lin(a, ncol_ext)
    a_out = _extrap2d_row_lin(a_.transpose(), nrow_ext).transpose()

    return a_out


def extrap2d_loglin(a, n_ext):
    """Extrapolate a 2-d array log-bilinearly.

    Parameters
    ----------
    a : array of float
        Input 2-d array.
    n_ext : int or tuple of int
        Number of extra elements on either side of each row and/or column.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 2-d array.

    """
    a = _check_2d_array(a, check_epsign=True)

    nrow_ext, ncol_ext = \
        n_ext if isinstance(n_ext, Sequence) else (n_ext, n_ext)

    # Extrapolate horizontally then vertically.
    a_ = _extrap2d_row_loglin(a, ncol_ext)
    a_out = _extrap2d_row_loglin(a_.transpose(), nrow_ext).transpose()

    return a_out


def extrap2d_loglin_oscil(a, n_ext):
    """Extrapolate a 2-d array with oscillatory behaviour log-bilinearly.

    Parameters
    ----------
    a : array of float
        Input 2-d array.
    n_ext : int or tuple of int
        Number of extra elements on either side of each row and/or column.

    Returns
    -------
    a_out : :class:`numpy.ndarray`
        Extrapolated 2-d array.

    """
    a = _check_2d_array(a)

    nrow_ext, ncol_ext = \
        n_ext if isinstance(n_ext, Sequence) else (n_ext, n_ext)

    # Extrapolate horizontally then vertically.
    a_ = _extrap2d_row_loglin_oscil(a, ncol_ext)
    a_out = _extrap2d_row_loglin_oscil(a_.transpose(), nrow_ext).transpose()

    return a_out


# ========================================================================
# Reshaping
# ========================================================================

def reshape_threept_datatab(paired_coords, flat_multipole, shape='triu'):
    """Reshape flattened three-point clustering measurements into 2-d
    arrays.

    Parameters
    ----------
    paired_coords : 2-d array of float
        Paired coordinate points, e.g. rows of :math:`(k_1, k_2)` values.
    flat_multipole : 1-d array of float
        Flattened multipole values corresponding to the
        paired coordinates, assuming the two coordinate variables
        are sampled at the same points.
    shape : {'triu', 'full'}, optional
        Shape of the input array without flattening.  If 'triu' (default),
        it corresponds to an upper triangular matrix and the output
        2-d array is symmetric; if 'full', it corresponds to a
        full matrix.

    Returns
    -------
    common_coords : 1-d array of float
        Common coordinate points as a vector.
    multipole : 2-d array of float
        Multipole values as a symmetric matrix.

    Raises
    ------
    ValueError
        If `paired_coords` is not a 2-d array.
    ValueError
        If `paired_coords` and `flat_multipole` have different lengths.
    ValueError
        If `paired_coords` and `flat_multipole` do not correspond to an
        upper triangular matrix.

    """
    # Check dimensions and reshape.
    if np.array(paired_coords).ndim != 2:
        raise ValueError("Paired coordinate points must be a 2-d array.")
    if (flat_len := len(paired_coords)) != len(flat_multipole):
        raise ValueError(
            "Paired coordinate points and flattened multipole "
            "array must have the same length."
        )

    if shape == 'triu':
        # Determine data dimensions.
        nbins = (np.sqrt(8*flat_len + 1) - 1) / 2.
        if not np.isclose(nbins, int(nbins)):
            raise ValueError(
                "Input arrays do not correspond to an upper triangular matrix."
            )
        nbins = int(nbins)

        # Reshape the flattened multipole array into a symmetric matrix.
        common_coords = paired_coords[:nbins, 1]

        triu_indices = np.triu_indices(nbins)

        multipole = np.zeros((nbins, nbins))
        multipole[triu_indices] = flat_multipole
        multipole.T[triu_indices] = flat_multipole
    elif shape == 'full':
        # Determine data dimensions.
        nbins = np.sqrt(flat_len)
        if not np.isclose(nbins, int(nbins)):
            raise ValueError(
                "Input arrays do not correspond to a square matrix."
            )
        nbins = int(nbins)

        # Reshape the flattened multipole array into a full matrix.
        common_coords = paired_coords[:nbins, 1]
        multipole = flat_multipole.reshape((nbins, nbins))
    else:
        raise ValueError(
            f"Invalid shape specified: '{shape}' not 'triu' or 'full'."
        )

    return common_coords, multipole



"""Test :mod:`~triumvirate.threept`.

"""
import numpy as np
import pytest

from triumvirate.threept import (
    compute_3pcf,
    compute_3pcf_in_gpp_box,
    compute_3pcf_window,
    compute_bispec,
    compute_bispec_in_gpp_box
)


@pytest.mark.slow
@pytest.mark.parametrize(
    "degrees, form, idx_bin",
    [
        ((0, 0, 0), 'diag', None),
        ((2, 0, 2), 'diag', None),
        ((0, 0, 0), 'row', 0),
    ]
)
def test_compute_bispec(degrees, form, idx_bin,
                        test_data_catalogue, test_rand_catalogue,
                        test_binning_fourier,
                        test_paramset,
                        test_logger,
                        test_stats_dir):

    degree_tag = ''.join(map(str, degrees))
    if form == 'off-diag':
        form_tag = 'offdiag{:d}'.format(idx_bin)
    elif form == 'row':
        form_tag = 'row{:d}'.format(idx_bin)
    else:
        form_tag = form

    measurements = compute_bispec(
        test_data_catalogue, test_rand_catalogue,
        degrees=degrees,
        binning=test_binning_fourier,
        form=form,
        idx_bin=idx_bin,
        paramset=test_paramset,
        logger=test_logger
    )

    measurements_ext = np.loadtxt(
        test_stats_dir/f"bk{degree_tag}_{form_tag}_lpp.txt", unpack=True
    )

    assert np.allclose(measurements['k1_bin'], measurements_ext[0]), \
        "Measurement bins do not match."
    assert np.allclose(measurements['k1_eff'], measurements_ext[1]), \
        "Measured coordinates do not match."
    assert np.allclose(measurements['nmodes_1'], measurements_ext[2]), \
        "Measured mode counts do not match."
    assert np.allclose(measurements['k2_bin'], measurements_ext[3]), \
        "Measurement bins do not match."
    assert np.allclose(measurements['k2_eff'], measurements_ext[4]), \
        "Measured coordinates do not match."
    assert np.allclose(measurements['nmodes_2'], measurements_ext[5]), \
        "Measured mode counts do not match."
    assert np.allclose(
        measurements['bk_raw'],
        measurements_ext[-4] + 1j * measurements_ext[-3]
    ), "Measured raw statistics do not match."
    assert np.allclose(
        measurements['bk_shot'],
        measurements_ext[-2] + 1j * measurements_ext[-1],
        atol=1.e-6
    ), "Measured shot noise contributions do not match."


@pytest.mark.slow
@pytest.mark.parametrize(
    "degrees, form, idx_bin",
    [
        ((0, 0, 0), 'diag', None),
        ((2, 0, 2), 'diag', None),
        ((0, 0, 0), 'row', 0),
    ]
)
def test_compute_3pcf(degrees, form, idx_bin,
                      test_data_catalogue, test_rand_catalogue,
                      test_binning_config,
                      test_paramset,
                      test_logger,
                      test_stats_dir):

    degree_tag = ''.join(map(str, degrees))
    if form == 'off-diag':
        form_tag = 'offdiag{:d}'.format(idx_bin)
    elif form == 'row':
        form_tag = 'row{:d}'.format(idx_bin)
    else:
        form_tag = form

    measurements = compute_3pcf(
        test_data_catalogue, test_rand_catalogue,
        degrees=degrees,
        binning=test_binning_config,
        form=form,
        idx_bin=idx_bin,
        paramset=test_paramset,
        logger=test_logger
    )
    measurements_ext = np.loadtxt(
        test_stats_dir/f"zeta{degree_tag}_{form_tag}_lpp.txt", unpack=True
    )

    assert np.allclose(
        measurements['r1_bin'], measurements_ext[0]
    ), "Measurement bins do not match."
    assert np.allclose(
        measurements['r1_eff'], measurements_ext[1]
    ), "Measured coordinates do not match."
    assert np.allclose(
        measurements['npairs_1'], measurements_ext[2]
    ), "Measured pair counts do not match."
    assert np.allclose(
        measurements['r2_bin'], measurements_ext[3]
    ), "Measurement bins do not match."
    assert np.allclose(
        measurements['r2_eff'], measurements_ext[4]
    ), "Measured coordinates do not match."
    assert np.allclose(
        measurements['npairs_2'], measurements_ext[5]
    ), "Measured pair counts do not match."
    assert np.allclose(
        measurements['zeta_raw'],
        measurements_ext[-4] + 1j * measurements_ext[-3]
    ), "Measured raw statistics do not match."
    assert np.allclose(
        measurements['zeta_shot'],
        measurements_ext[-2] + 1j * measurements_ext[-1],
        atol=1.e-6
    ), "Measured shot noise contributions do not match."


@pytest.mark.slow
@pytest.mark.parametrize(
    "degrees, form, idx_bin",
    [
        ((0, 0, 0), 'diag', None),
        ((2, 0, 2), 'diag', None),
        ((0, 0, 0), 'row', 0),
    ]
)
def test_compute_bispec_in_gpp_box(degrees, form, idx_bin,
                                   test_data_catalogue,
                                   test_binning_fourier,
                                   test_paramset,
                                   test_logger,
                                   test_stats_dir):

    degree_tag = ''.join(map(str, degrees))
    if form == 'off-diag':
        form_tag = 'offdiag{:d}'.format(idx_bin)
    elif form == 'row':
        form_tag = 'row{:d}'.format(idx_bin)
    else:
        form_tag = form

    measurements = compute_bispec_in_gpp_box(
        test_data_catalogue,
        degrees=degrees,
        binning=test_binning_fourier,
        form=form,
        idx_bin=idx_bin,
        paramset=test_paramset,
        logger=test_logger
    )
    measurements_ext = np.loadtxt(
        test_stats_dir/f"bk{degree_tag}_{form_tag}_gpp.txt", unpack=True
    )

    assert np.allclose(measurements['k1_bin'], measurements_ext[0]), \
        "Measurement bins do not match."
    assert np.allclose(measurements['k1_eff'], measurements_ext[1]), \
        "Measured coordinates do not match."
    assert np.allclose(measurements['nmodes_1'], measurements_ext[2]), \
        "Measured mode counts do not match."
    assert np.allclose(measurements['k2_bin'], measurements_ext[3]), \
        "Measurement bins do not match."
    assert np.allclose(measurements['k2_eff'], measurements_ext[4]), \
        "Measured coordinates do not match."
    assert np.allclose(measurements['nmodes_2'], measurements_ext[5]), \
        "Measured mode counts do not match."
    assert np.allclose(
        measurements['bk_raw'],
        measurements_ext[-4] + 1j * measurements_ext[-3]
    ), "Measured raw statistics do not match."
    assert np.allclose(
        measurements['bk_shot'],
        measurements_ext[-2] + 1j * measurements_ext[-1],
        atol=1.e-6
    ), "Measured shot noise contributions do not match."


@pytest.mark.slow
@pytest.mark.parametrize(
    "degrees, form, idx_bin",
    [
        ((0, 0, 0), 'diag', None),
        ((2, 0, 2), 'diag', None),
        ((0, 0, 0), 'row', 0),
    ]
)
def test_compute_3pcf_in_gpp_box(degrees, form, idx_bin,
                                 test_data_catalogue,
                                 test_binning_config,
                                 test_paramset,
                                 test_logger,
                                 test_stats_dir):

    degree_tag = ''.join(map(str, degrees))
    if form == 'off-diag':
        form_tag = 'offdiag{:d}'.format(idx_bin)
    elif form == 'row':
        form_tag = 'row{:d}'.format(idx_bin)
    else:
        form_tag = form

    measurements = compute_3pcf_in_gpp_box(
        test_data_catalogue,
        degrees=degrees,
        binning=test_binning_config,
        form=form,
        idx_bin=idx_bin,
        paramset=test_paramset,
        logger=test_logger
    )
    measurements_ext = np.loadtxt(
        test_stats_dir/f"zeta{degree_tag}_{form_tag}_gpp.txt", unpack=True
    )

    assert np.allclose(
        measurements['r1_bin'], measurements_ext[0]
    ), "Measurement bins do not match."
    assert np.allclose(
        measurements['r1_eff'], measurements_ext[1]
    ), "Measured coordinates do not match."
    assert np.allclose(
        measurements['npairs_1'], measurements_ext[2]
    ), "Measured pair counts do not match."
    assert np.allclose(
        measurements['r2_bin'], measurements_ext[3]
    ), "Measurement bins do not match."
    assert np.allclose(
        measurements['r2_eff'], measurements_ext[4]
    ), "Measured coordinates do not match."
    assert np.allclose(
        measurements['npairs_2'], measurements_ext[5]
    ), "Measured pair counts do not match."
    assert np.allclose(
        measurements['zeta_raw'],
        measurements_ext[-4] + 1j * measurements_ext[-3]
    ), "Measured raw statistics do not match."
    assert np.allclose(
        measurements['zeta_shot'],
        measurements_ext[-2] + 1j * measurements_ext[-1],
        atol=1.e-6
    ), "Measured shot noise contributions do not match."


@pytest.mark.slow
@pytest.mark.parametrize(
    "degrees, form, idx_bin",
    [
        ((0, 0, 0), 'diag', None),
        ((2, 0, 2), 'diag', None),
        ((0, 0, 0), 'row', 0),
    ]
)
def test_compute_3pcf_window(degrees, form, idx_bin,
                             test_rand_catalogue,
                             test_binning_config,
                             test_paramset,
                             test_logger,
                             test_stats_dir):

    degree_tag = ''.join(map(str, degrees))
    if form == 'off-diag':
        form_tag = 'offdiag{:d}'.format(idx_bin)
    elif form == 'row':
        form_tag = 'row{:d}'.format(idx_bin)
    else:
        form_tag = form

    measurements = compute_3pcf_window(
        test_rand_catalogue,
        degrees=degrees,
        binning=test_binning_config,
        form=form,
        idx_bin=idx_bin,
        paramset=test_paramset,
        logger=test_logger
    )
    measurements_ext = np.loadtxt(
        test_stats_dir/f"zetaw{degree_tag}_{form_tag}.txt", unpack=True
    )

    assert np.allclose(
        measurements['r1_bin'], measurements_ext[0]
    ), "Measurement bins do not match."
    assert np.allclose(
        measurements['r1_eff'], measurements_ext[1]
    ), "Measured coordinates do not match."
    assert np.allclose(
        measurements['npairs_1'], measurements_ext[2]
    ), "Measured pair counts do not match."
    assert np.allclose(
        measurements['r2_bin'], measurements_ext[3]
    ), "Measurement bins do not match."
    assert np.allclose(
        measurements['r2_eff'], measurements_ext[4]
    ), "Measured coordinates do not match."
    assert np.allclose(
        measurements['npairs_2'], measurements_ext[5]
    ), "Measured pair counts do not match."
    assert np.allclose(
        measurements['zeta_raw'],
        measurements_ext[-4] + 1j * measurements_ext[-3]
    ), "Measured raw statistics do not match."
    assert np.allclose(
        measurements['zeta_shot'],
        measurements_ext[-2] + 1j * measurements_ext[-1],
        atol=1.e-6
    ), "Measured shot noise contributions do not match."



"""
MPI Tools (:mod:`~triumvirate._mpitools`)
=========================================

MPI-based parallelisation tools.

.. autosummary::
    allocate_tasks
    distribute_tasks

"""  # numpydoc ignore=SS01
import warnings

import numpy as np


def allocate_tasks(task_total, proc_total):
    """Allocate tasks to processes.

    This assigns ``tasks[i]`` tasks to the rank-``i`` process such that
    the total number of tasks, `task_total`, is shared amongst all
    `proc_total` processes.

    Parameters
    ----------
    task_total : int
        Total number of tasks.
    proc_total : int
        Total number of processes.

    Returns
    -------
    ntasks : list of int
        Number of tasks for each process.

    Raises
    ------
    TypeError
        When `task_total` or `proc_total` is not an integer.
    ValueError
        When `task_total` or `proc_total` is not positive.

    """
    if not isinstance(task_total, (int, np.integer)) \
            or not isinstance(proc_total, int):
        raise TypeError(
            "Both `task_total` and `proc_total` must be integers: "
            f"got {type(task_total)=} and {type(proc_total)=}"
        )
    if task_total <= 0 or proc_total <= 0:
        raise ValueError(
            "Both `task_total` and `proc_total` must be positive."
        )

    ntask_toassign, nproc_toassign, ntasks = task_total, proc_total, []
    while ntask_toassign > 0:
        ntask_assigned = ntask_toassign // nproc_toassign
        ntasks.append(ntask_assigned)
        ntask_toassign -= ntask_assigned
        nproc_toassign -= 1

    return ntasks


def distribute_tasks(ntasks=None, task_total=None, proc_total=None):
    """Distribute segments of tasks to each process.

    The number of tasks each process receives is determined by
    :func:`allocate_tasks` if `tasks` is `None`.  The rank-``i`` process
    receives ``ntasks[i]`` tasks in the range given by ``segments[i]``.

    Parameters
    ----------
    ntasks : list of int, optional
        Number of tasks each process receives.  If `None` (default),
        `task_total` and `proc_total` must be provided; otherwise,
        `task_total` and `proc_total` are both ignored.
    task_total : int, optional
        Total number of tasks (default is `None`).
    proc_total : int, optional
        Total number of processes (default is `None`).

    Returns
    -------
    segments : list of slice
        Index slices corresponding to the segment of tasks that each
        process should receive.

    """
    ntasks = ntasks if ntasks else allocate_tasks(task_total, proc_total)
    proc_total = proc_total if proc_total else len(ntasks)

    breakpoints = np.insert(np.cumsum(ntasks), 0, values=0)
    segments = [
        slice(breakpoints[rank], breakpoints[rank + 1])
        for rank in range(proc_total)
    ]

    return segments


def restore_warnings(captured_warnings, unique=True, show=True,
                     comm=None, comm_root=0):
    """Emit captured warnings.

    Parameters
    ----------
    captured_warnings : list of :class:`warnings.WarningMessage`
        List of recorded warnings as returned by
        ``warnings.catch_warnings(record=True)``.
    unique : bool, optional
        If `True` (default), only emit unique warnings; otherwise, emit
        all warnings.
    show : bool, optional
        If `True` (default), emit the warnings; otherwise, do not emit
        the warnings.
    comm : :class:`mpi4py.MPI.Comm`, optional
        MPI communicator (default is `None`).
    comm_root : int, optional
        Root process rank of the MPI communicator (default is 0).
        Ignored if `comm` is `None`.

    Returns
    -------
    restored_warnings : list of :class:`warnings.WarningMessage`
        Restored warnings.

    """
    if unique:
        restored_warnings = []
        seen_warnings = set()
        for record in captured_warnings:
            record_tuple = (
                record.message, record.category, record.filename, record.lineno
            )
            if record_tuple not in seen_warnings:
                restored_warnings.append(record)
                seen_warnings.add(record_tuple)
    else:
        restored_warnings = captured_warnings

    if show and (comm is None or comm.rank == comm_root):
        for record in restored_warnings:
            warnings.showwarning(
                record.message, record.category,
                record.filename, record.lineno,
                file=record.file, line=record.line
            )

    return restored_warnings
